#! /usr/bin/env node
'use strict';

var fs = require('fs');
var inquirer = require('inquirer');
var log4js = require('log4js');
var nconf = require('nconf');
var request = require('request').defaults({ jar: true });
var util = require('util');
var when = require('when');

var logger = log4js.getLogger('create');

var configFilePath = '.pullrequest-config.json';
var issueEndpoint = '/rest/issue/';
// This is the publicly shareable API key, safe to use for the infrequent personal use.
// If you are using this module to submit a massive number of PRs, you need to get registered API key from Giphy.
var giphyPublicBetaKey = 'dc6zaTOxFJmzC';
var giphyRandomEndpoint = 'http://api.giphy.com/v1/gifs/random';

ensureConfigExists(configFilePath);

promptForPasscode({}).then(promptForYouTrackSettings).then(validateYouTrackCredentials).then(promptForGitHubCredentials).then(validateGitHubCredentials).then(promptForRepo).then(promptForHead).then(promptForBase).then(promptForIssueID).then(getIssueTitle).then(promptForDescription).then(promptForSummaryBullets).then(promptForDetailBullets).then(getGIF).then(submitPR);

function loadConfig(passcode, configFilePath) {
  nconf.argv().env().file({
    file: configFilePath,
    secure: {
      secret: passcode,
      alg: 'aes-256-ctr'
    }
  });
}

function promptForPasscode(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'password',
      name: 'passcode',
      message: 'Enter your passcode for settings stored in .pullrequest-config.json:',
      validate: function validate(value) {
        return !!value;
      }
    }], function (answers) {
      loadConfig(answers.passcode, configFilePath);
      resolve(state);
    });
  });
}

function promptForYouTrackSettings(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'youTrackURL',
      message: 'Enter the URL to your YouTrack dashboard (e.g. URL ends in /dashboard):',
      when: function when() {
        return !nconf.get('youtrack:url');
      },
      validate: function validate(value) {
        return value.indexOf('/dashboard') !== -1 && value.indexOf('https://') === 0;
      },
      filter: function filter(value) {
        return value.substr(0, value.indexOf('/dashboard'));
      }
    }, {
      type: 'input',
      name: 'youTrackLogin',
      message: 'Enter your YouTrack login:',
      when: function when() {
        return !nconf.get('youtrack:login');
      },
      validate: function validate(value) {
        return !!value;
      }
    }, {
      type: 'password',
      name: 'youTrackPassword',
      message: 'Enter your YouTrack password:',
      when: function when() {
        return !nconf.get('youtrack:password');
      },
      validate: function validate(value) {
        return !!value;
      }
    }], function (answers) {
      if (Object.keys(answers).length) {
        updateStoredSettings(answers);
      }
      state.url = answers.url || nconf.get('youtrack:url');
      state.youTrackLogin = answers.youTrackLogin || nconf.get('youtrack:login');
      state.youTrackPassword = answers.youTrackPassword || nconf.get('youtrack:password');
      resolve(state);
    });
  });
}

function validateYouTrackCredentials(state) {
  return when.promise(function (resolve, reject) {
    request.post({
      url: state.url + '/rest/user/login',
      form: {
        login: state.youTrackLogin,
        password: state.youTrackPassword
      }
    }, function (err, res) {
      if (err || res.statusCode !== 200) {
        logger.error('Unable to login: ' + err + ' status: ' + (res && res.statusCode));
        nconf.clear('youtrack:url');
        nconf.clear('youtrack:login');
        nconf.clear('youtrack:password');
        reject(state);
      } else {
        resolve(state);
      }
    });
  });
}

function promptForGitHubCredentials(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'gitHubLogin',
      message: 'Enter your GitHub user ID:',
      when: function when() {
        return !nconf.get('github:login');
      },
      validate: function validate(value) {
        return !!value;
      }
    }, {
      type: 'password',
      name: 'gitHubPassword',
      message: 'Enter a GitHub personal access token with \'repo\' access (created at ' + ' https://github.com/settings/tokens/new):',
      when: function when() {
        return !nconf.get('github:password');
      },
      validate: function validate(value) {
        return !!value;
      }
    }], function (answers) {
      updateStoredSettings(answers);
      state.gitHubLogin = answers.gitHubLogin || nconf.get('github:login');
      state.gitHubPassword = answers.gitHubPassword || nconf.get('github:password');
      resolve(state);
    });
  });
}

function validateGitHubCredentials(state) {
  return when.promise(function (resolve, reject) {
    request.get({
      url: 'https://api.github.com/user',
      headers: {
        Accept: 'application/vnd.github.v3+json',
        Authorization: 'Basic ' + new Buffer(state.gitHubLogin + ':' + state.gitHubPassword, 'ascii').toString('base64'),
        'user-agent': 'NodeJS HTTP Client'
      }
    }, function (err, res) {
      if (err) {
        logger.error('Credentials are invalid: ' + err + ' res: ' + util.inspect(res));
        reject(state);
      } else {
        resolve(state);
      }
    });
  });
}

function promptForIssueID(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'issueID',
      message: 'Enter the YouTrack issue ID to associate with this PR:',
      validate: function validate(value) {
        return !value || !!value.match(/[A-Za-z0-9]+\-\d+/);
      }
    }], function (answers) {
      state.issueID = answers.issueID;
      resolve(state);
    });
  });
}

function getIssueTitle(state) {
  return when.promise(function (resolve, reject) {
    if (state.issueID) {
      request.get({
        url: nconf.get('youtrack:url') + issueEndpoint + state.issueID,
        headers: { Accept: 'application/json' }
      }, function (err, res, body) {
        if (err || res.statusCode !== 200) {
          logger.error('Unable to get issue information: ' + err + ' status: ' + (res && res.statusCode));
          reject(state);
        } else {
          var data = JSON.parse(body);
          var id = data.id;
          state.title = id + ' - ' + data.field.filter(function (value) {
            return value.name === 'summary';
          })[0].value;
          resolve(state);
        }
      });
    } else {
      resolve(state);
    }
  });
}

function promptForDescription(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'title',
      message: 'Title for the PR:',
      validate: function validate(value) {
        return !!value;
      },
      default: state.title
    }, {
      type: 'confirm',
      name: 'hasReviewers',
      message: 'Are there required reviewers?',
      default: true
    }, {
      type: 'input',
      name: 'reviewers',
      message: 'Enter required reviewers, comma separated:',
      when: function when(answers) {
        return answers.hasReviewers;
      },
      default: nconf.get('reviewers')
    }, {
      type: 'input',
      name: 'buddyURL',
      message: 'Enter the buddy build base URL (e.g. everything before the ID):',
      when: function when() {
        return !nconf.get('buddyURL');
      },
      validate: function validate(value) {
        return value.indexOf('http') === 0 && value.charAt(value.length - 1) === '/';
      }
    }, {
      type: 'input',
      name: 'buddyID',
      message: 'Enter the buddy build ID:',
      when: function when(answers) {
        return !!nconf.get('buddyURL') || !!answers.buddyURL;
      },
      validate: function validate(value) {
        return !value || !!value.match(/^\d+$/);
      }
    }, {
      type: 'confirm',
      name: 'green',
      message: 'Is the buddy build green?',
      when: function when(answers) {
        return !!answers.buddyID;
      },
      default: true
    }, {
      type: 'input',
      name: 'giphyTag',
      message: 'What is the giphy tag?'
    }], function (answers) {
      state.title = answers.title;
      state.reviewers = answers.reviewers, state.buddyURL = answers.buddyURL || nconf.get('buddyURL');
      state.buddyID = answers.buddyID;
      state.green = answers.green;
      state.giphyTag = answers.giphyTag;
      updateStoredSettings(answers);
      resolve(state);
    });
  });
}

function promptForSummaryBullets(state) {
  state.summaryBullets = [];
  var message = 'Enter a summary bullet:';
  return when.promise(function (resolve) {
    var prompt = function prompt() {
      inquirer.prompt([{
        type: 'input',
        name: 'summaryBullet',
        message: message,
        validate: function validate(value) {
          return !!value || state.summaryBullets.length > 0;
        }
      }], function (answers) {
        if (answers.summaryBullet) {
          state.summaryBullets.push(answers.summaryBullet);
          message = 'Enter another summary bullet (optional):';
          process.nextTick(prompt);
        } else {
          resolve(state);
        }
      });
    };
    prompt();
  });
}

function promptForDetailBullets(state) {
  state.detailBullets = [];
  var message = 'Enter a detail bullet (optional):';
  return when.promise(function (resolve) {
    var prompt = function prompt() {
      inquirer.prompt([{
        type: 'input',
        name: 'detailBullet',
        message: message
      }], function (answers) {
        if (answers.detailBullet) {
          state.detailBullets.push(answers.detailBullet);
          message = 'Enter another detail bullet (optional):';
          process.nextTick(prompt);
        } else {
          resolve(state);
        }
      });
    };
    prompt();
  });
}

function getGIF(state) {
  return when.promise(function (resolve, reject) {
    request.get({
      url: giphyRandomEndpoint,
      headers: {
        'Accept': 'application/json',
        'user-agent': 'NodeJS HTTP Client'
      },
      qs: {
        tag: state.giphyTag,
        rating: 'pg-13',
        api_key: giphyPublicBetaKey
      }
    }, function (err, res) {
      if (err || res && res.statusCode !== 200) {
        logger.error('Unable to get GIF: ' + err + ' res: ' + util.inspect(res));
        reject();
      } else {
        state.gif = JSON.parse(res.body).data.image_original_url;
        resolve(state);
      }
    });
  });
}

function promptForRepo(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'gitHubRepoOwner',
      message: 'Enter the GitHub repository owner:',
      when: function when() {
        return !nconf.get('github:repo:owner');
      },
      validate: function validate(value) {
        return !!value;
      }
    }, {
      type: 'input',
      name: 'gitHubRepoName',
      message: 'Enter the GitHub repository name:',
      when: function when() {
        return !nconf.get('github:repo:name');
      },
      validate: function validate(value) {
        return !!value;
      }
    }], function (answers) {
      state.gitHubRepoOwner = answers.gitHubRepoOwner;
      state.gitHubRepoName = answers.gitHubRepoName;
      updateStoredSettings(answers);
      resolve(state);
    });
  });
}

function promptForHead(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'head',
      message: 'Enter the name of your branch:',
      validate: function validate(value) {
        return !!value;
      }
    }], function (answers) {
      state.head = answers.head;
      resolve(state);
    });
  });
}

function promptForBase(state) {
  return when.promise(function (resolve) {
    inquirer.prompt([{
      type: 'input',
      name: 'base',
      message: 'Enter the name of the branch to merge into:',
      default: 'master',
      validate: function validate(value) {
        return !!value;
      }
    }], function (answers) {
      state.base = answers.base;
      resolve(state);
    });
  });
}

var prTemplate = '### SUMMARY\n' + '\n' + '{summaryBullets}\n' + '\n' + '{requiredReviewersSection}' + // no \n since section is optional
'{detailSection}' + // no \n since section is optional
'{buddySection}' + // no \n since section is optional
'### PIC\n' + '\n' + '![]({gif})';
var requiredReviewersTemplate = '### REQUIRED REVIEWERS\n' + '\n' + '{requiredReviewers}\n' + '\n';
var buddyTemplate = '### BUDDY\n' + '\n' + 'Buddy Build: {buddyURL}{buddyID}\n' + '- [{green}] Green\n' + '\n';
var detailTemplate = '### DETAIL\n' + '\n' + '{detailBullets}\n' + '\n';

function createPRDescription(state) {
  var prBody = prTemplate;
  var summaryBullets = '';
  state.summaryBullets.forEach(function (summaryBullet) {
    summaryBullets += '* ' + summaryBullet + '\n';
  });
  prBody = prBody.replace('{summaryBullets}', summaryBullets);
  if (state.detailBullets.length) {
    prBody.replace('{detailSection}', detailTemplate);
    var detailBullets = '';
    state.detailBullets.forEach(function (detailBullet) {
      detailBullets += '* ' + detailBullet + '\n';
    });
    prBody = prBody.replace('{detailBullets}', detailBullets);
  } else {
    prBody = prBody.replace('{detailSection}', '');
  }
  if (state.reviewers) {
    prBody = prBody.replace('{requiredReviewersSection}', requiredReviewersTemplate);
    var reviewers = '';
    state.reviewers.split(',').forEach(function (reviewer) {
      reviewers += '- [ ] ' + reviewer + '\n';
    });
    prBody = prBody.replace('{requiredReviewers}', reviewers);
  } else {
    prBody = prBody.replace('{requiredReviewersSection}', '');
  }
  if (state.buddyID) {
    prBody = prBody.replace('{buddySection}', buddyTemplate);
    prBody = prBody.replace('{green}', state.green ? 'x' : ' ');
    prBody = prBody.replace('{buddyID}', state.buddyID);
    prBody = prBody.replace('{buddyURL}', state.buddyURL);
  } else {
    prBody = prBody.replace('{buddySection}', '');
  }
  prBody = prBody.replace('{gif}', state.gif);
  return prBody;
}

function submitPR(state) {
  var body = createPRDescription(state);
  var auth = new Buffer(nconf.get('github:login') + ':' + nconf.get('github:password'), 'ascii').toString('base64');

  request.post({
    url: 'https://api.github.com/repos/' + nconf.get('github:repo:owner') + '/' + nconf.get('github:repo:name') + '/pulls',
    headers: {
      'Accept': 'application/vnd.github.v3+json',
      'Authorization': 'Basic ' + auth,
      'user-agent': 'NodeJS HTTP Client'
    },
    body: JSON.stringify({
      'title': state.title,
      'body': body,
      'head': state.head,
      'base': state.base
    })
  }, function (err, res) {
    if (err) {
      logger.error('Unable to submit pull request: ' + err + ' res: ' + util.inspect(res));
    } else {
      logger.info('Success! Status: ' + (res && res.statusCode));
    }
  });
}

function updateStoredSettings(answers) {
  setIfProvided('youtrack:url', answers.youTrackURL);
  setIfProvided('youtrack:login', answers.youTrackLogin);
  setIfProvided('youtrack:password', answers.youTrackPassword);
  setIfProvided('github:login', answers.gitHubLogin);
  setIfProvided('github:password', answers.gitHubPassword);
  setIfProvided('buddyURL', answers.buddyURL);
  setIfProvided('github:repo:owner', answers.gitHubRepoOwner);
  setIfProvided('github:repo:name', answers.gitHubRepoName);
  setIfProvided('reviewers', answers.reviewers);
  nconf.save(function (err) {
    if (err) {
      logger.error('Unable to save config at: ' + configFilePath + ' error: ' + err);
    } else {
      // logger.debug('Defaults saved in: ' + configFilePath);
    }
  });
}

function setIfProvided(key, value) {
  if (value) {
    nconf.set(key, value);
  }
}

function ensureConfigExists(configFilePath) {
  if (!fs.existsSync(configFilePath)) {
    fs.writeFileSync(configFilePath, '{}');
  }
}

/* TODO
Change status of task
Open up compare for preview on github
Get labels and prompt to choose (remembering defaults)
automatically check to see if you have latest version and ask to update if not
Record on YouTrack of PR that is opened
Implement merge, which will automatically copy description text and past into merge message.
Provide option to close YouTrack item on merge.
Ability to automatically clone a YouTrack item to put remaining work into it.
*/